%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File name: RPEM.m
%
% Description: MIMO RPEM algorithm for general nonlinear
%              identification. Include noise covariance
%              estimation.
%
% Inputs: Input signals, dimensions and output signals.
%
% Outputs: Estimated parameters.
%
% Author and copyright: Torbjrn Wigren
%                       Systems and Control
%                       Department of Information Technology
%                       Uppsala University
%                       Uppsala
%                       Sweden
%
% Revision log: Created 2003-04-27.
%			    Fixed output equation error, time index of u
%               changed from t to t+1, 2005-08-23.
%
% Firm revision: revision 1.0, 2004-12-23.
%                revision 2.0, 2005-08-23.
%                revision 3.0, 2010-01-14.
%                revision 4.0, 2010-09-06.	
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Dimensions
%

%  PolynomialOrders=zeros(nx+nu,1);
%
%  Example:
%
%  polynomialOrders=[3 1 3]';
%  notUsedIndices=[
%   1 1 1;
%   2 1 1;
%   3 1 1];
%
%  Note: First index corresponds to x_1 (slowest variation), the
%        following to x_2 and so on.
%

%
% Generate indices to sum over
%

if ~exist('theta_0_new')
    'Input polynomialOrders and notUsedIndices'
    keyboard;
        allIndices=GenerateIndices(polynomialOrders,notUsedIndices)
    dimTheta=size(allIndices,1);
    LinearizedPoleRadii=eye(nx);
        'Input method'
        keyboard;
    while any(LinearizedPoleRadii>stabilityLimit)
        'Input theta_0'
        keyboard;
        if strcmp(method,'Euler')
            SM=eye(nx)+scalingTs*Ts*dfdx_m(x_m_0,u(:,1),theta_0,allIndices);
        elseif strcmp(method,'midpoint')
            SM=eye(nx)+scalingTs*Ts*dfdx_m(x_m_0+0.5*scalingTs*Ts*...
                f_m(x_m_0,u(:,1),theta_0,allIndices),u(:,1),theta_0,allIndices);
        end
        LinearizedPoleRadii=abs(eig(SM));
    end
end
if exist('theta_0_new')
        allIndices=GenerateIndices(polynomialOrders,notUsedIndices);
    dimTheta=size(allIndices,1);
    theta_0=theta_0_new;
end

%
% Remaining initial values
%

Gamma_0=initialNoiseVariance*eye(size(y,1),size(y,1));
R_0=scaleFactorR*eye(dimTheta,dimTheta);
psi_l_0=zeros(dimTheta,nx);

% Storage variables

x_m_save=zeros(nx,ceil(N/downSampling)+1);
y_m_save=zeros(ny,ceil(N/downSampling)+1);
perr_save=zeros(ny,ceil(N/downSampling));
theta_save=zeros(dimTheta,ceil(N/downSampling)+1);
eig_R_save=zeros(dimTheta,ceil(N/downSampling)+1);
condition_save=zeros(1,ceil(N/downSampling)+1);

%
% Algorithmic loop
%

% Initiation at t=0...

mu=mu_0;
y_m=y_m_0;
Gamma=Gamma_0;
psi_l=psi_l_0;
x_m=x_m_0;
R=R_0;
theta=theta_0;

x_m_save(:,1)=x_m_0;
y_m_save(:,1)=y_m_0;
theta_save(:,1)=theta_0;

percentReady=0;

for t=1:N
    
    if round(t*100/N)>percentReady
        percentReady=round(t*100/N);
    end
    
    %
    % Calculate mu(t)
    %
    
    mu=mu0*mu+1-mu0;
    
    %
    % Perform a stability check
    %
    if strcmp(method,'Euler')
        SM=eye(nx)+scalingTs*Ts*dfdx_m(x_m,u(:,t),theta,allIndices);
    elseif strcmp(method,'midpoint')
        SM = eye(nx)+scalingTs*Ts*dfdx_m(x_m+0.5*scalingTs*Ts*...
            f_m(x_m,u(:,t),theta,allIndices),u(:,t),theta,allIndices);
    end
    LinearizedPoles=eig(SM);
    if all(abs(LinearizedPoles)<stabilityLimit)
        stable=1;
    else
        stable=0;
    end
    
    %
    % Prediction error and Gamma update
    %
    
    perr=y(:,t)-y_m;
    Gamma=Gamma+(mu/(muFactor+t))*(perr*perr'-Gamma);
    
    %
    % Covariance update
    %
    
    R=R+(mu/(muFactor+t))*(psi_l*dhdx_m(x_m,u(:,t))'*inv(Gamma)*dhdx_m(x_m,u(:,t))*psi_l'-R);
    
    %
    % Parameter update
    %
    
    if stable
        theta=theta+(mu/(muFactor+t))*inv(R)*psi_l*dhdx_m(x_m,u(:,t))'*inv(Gamma)*perr;
    end
    
    %
    % Check which method is chosen by the user
    %
    
    if strcmp(method,'Euler')
        %
        % Gradient computation
        %
        
        psi_l = psi_l + scalingTs*Ts*...
            (psi_l*dfdx_m(x_m,u(:,t),theta,allIndices)'+...
            dfdtheta_m(x_m,u(:,t),theta,allIndices)');
        %
        % State update
        %
        
        x_m = x_m + scalingTs*Ts*f_m(x_m,u(:,t),theta,allIndices);
        
    elseif strcmp(method,'midpoint')
        %
        % Gradient computation
        %
        
        psi_l = psi_l + scalingTs*Ts*((psi_l + scalingTs...
            *Ts/2*psi_l*dfdx_m(x_m,u(:,t),theta,allIndices)' + ...
            scalingTs*Ts/2*dfdtheta_m(x_m,u(:,t),theta,allIndices)')...
            *dfdx_m((x_m + scalingTs*Ts/2*f_m(x_m,u(:,t)...
            ,theta,allIndices)),u(:,t),theta,allIndices )'+ ...
            (dfdtheta_m(x_m + scalingTs*Ts/2*f_m(x_m,u(:,t)...
            ,theta,allIndices),u(:,t),theta,allIndices))');
        
        %
        % State update
        %
        
        x_m = x_m + scalingTs*Ts*(f_m(x_m + scalingTs*Ts*0.5*f_m(x_m,u(:,t),...
            theta,allIndices),u(:,t),theta,allIndices));
    else
        error('Please make sure to choose a method between either "Euler" or "midpoint"');
    end
    
    %
    % Model output computation
    %
    
    if t<N
        y_m=h_m(x_m,u(:,t+1));
    end
    
    %
    % Check and save
    %
    
    if rem(t,downSampling)==0 % Time to save
        x_m_save(:,t/downSampling+1)=x_m;
        y_m_save(:,t/downSampling+1)=y_m;
        perr_save(:,t/downSampling)=perr;
        theta_save(:,t/downSampling+1)=theta;
        eig_R_save(:,t/downSampling+1)=eig(R);
        condition_save(:,t/downSampling+1)=cond(R);
    end
    
end

reInitiateTheta=theta;
reInitiateIndices=allIndices;

%
% Compute the re-scaled, original parameter vector
%

rescaledTheta=theta;
for i=1:dimTheta
    for j=1:nx
        thisSubIndex=allIndices(i,j);
        rescaledTheta(i,1)=rescaledTheta(i,1)*((1/scalingTs)^(thisSubIndex*(j-1)));
    end
end
rescaledTheta=rescaledTheta*(scalingTs)^nx;

%
% End of algorithmic loop
%

[theta rescaledTheta];

